import os
import sys
import platform
import re
import shutil
import subprocess
import importlib
import locale
import json
import yaml
import traceback
from math import floor
from traceback import format_exc
from typing import Optional, Any, Tuple, List, Dict, Union

import termcolor
from termcolor import colored
from questionary import Style

import dt_authentication
from dt_authentication import DuckietownToken

from dt_shell_cli import logger
from . import __version__
from .constants import BASH_COMPLETION_DIR, SHELL_LIB_DIR, DTShellConstants
from .exceptions import ShellInitException, RunCommandException

NOTSET = object()


cli_style = Style([
    ('qmark', 'fg:#673ab7 bold'),        # token in front of the question
    ('question', 'bold'),                # question text
    ('choice', 'fg:#fec20b bold'),       # a possible choice in select
    ('answer', 'fg:#fec20b bold'),       # submitted answer text behind the question
    ('pointer', 'fg:#673ab7 bold'),      # pointer used in select and checkbox prompts
    ('highlighted', 'fg:#673ab7 bold'),  # pointed-at choice in select and checkbox prompts
    ('selected', 'fg:#cc5454'),          # style for a selected item of a checkbox
    ('separator', 'fg:#cc5454'),         # separator in lists
    ('instruction', ''),                 # user instructions for select, rawselect, checkbox
    ('text', ''),                        # plain text
    ('disabled', 'fg:#bbbbbb italic')    # disabled choices for select and checkbox prompts
])


def indent(s: str, prefix: str, first: Optional[str] = None) -> str:
    s = str(s)
    assert isinstance(prefix, str)
    lines = s.split("\n")
    if not lines:
        return ""

    if first is None:
        first = prefix

    m = max(len(prefix), len(first))

    prefix = " " * (m - len(prefix)) + prefix
    first = " " * (m - len(first)) + first

    # different first prefix
    res = ["%s%s" % (prefix, line.rstrip()) for line in lines]
    res[0] = "%s%s" % (first, lines[0].rstrip())
    return "\n".join(res)


def raise_wrapped(etype, e, msg, compact=False, exc=None, **kwargs):
    """Raises an exception of type etype by wrapping
    another exception "e" with its backtrace and adding
    the objects in kwargs as formatted by format_obs.

    if compact = False, write the whole traceback, otherwise just str(e).

    exc = output of sys.exc_info()
    """

    e = raise_wrapped_make(etype, e, msg, compact=compact, **kwargs)

    #     if exc is not None:
    #         _, _, trace = exc
    #         raise etype, e.args, trace
    #     else:
    raise e


def raise_wrapped_make(etype, e, msg, compact=False, **kwargs):
    """Constructs the exception to be thrown by raise_wrapped()"""
    assert isinstance(e, BaseException), type(e)
    assert isinstance(msg, str), type(msg)
    s = msg

    import sys

    if sys.version_info[0] >= 3:
        es = str(e)
    else:
        if compact:
            es = str(e)
        else:
            es = format_exc()

    s += "\n" + indent(es.strip(), "| ")

    return etype(s)


def check_isinstance(ob, expected, **kwargs):
    if not isinstance(ob, expected):
        kwargs["object"] = ob
        raise_type_mismatch(ob, expected, **kwargs)


def raise_type_mismatch(ob, expected, **kwargs):
    """Raises an exception concerning ob having the wrong type."""
    e = "Object not of expected type:"
    e += "\n  expected: %s" % str(expected)
    e += "\n  obtained: %s" % str(type(ob))
    # e += '\n' + indent(format_obs(kwargs), ' ')
    raise ValueError(e)


SPACE_TAG = "SPACE_TAG"


def replace_spaces(x: str) -> str:
    return x.replace(" ", SPACE_TAG)


def undo_replace_spaces(x: str) -> str:
    return x.replace(SPACE_TAG, " ")


def href(x):
    return termcolor.colored(x, "blue", None, ["underline"])


def dark_yellow(x):
    return termcolor.colored(x, "yellow")


def dark(x):
    return termcolor.colored(x, attrs=["dark"])


def safe_pathname(s: str) -> str:
    return re.sub(r"[^\w\d-]", "_", s)


def run_cmd(cmd, print_output=False, suppress_errors=False):
    logger.debug("$ %s" % cmd)
    # spawn new process
    proc = subprocess.Popen(cmd, stderr=subprocess.PIPE, stdout=subprocess.PIPE)
    stdout, stderr = proc.communicate()
    stdout = stdout.decode("utf-8") if stdout else None
    stderr = stderr.decode("utf-8") if stderr else None
    returncode = proc.returncode
    # ---
    if returncode != 0 and not suppress_errors:
        msg = "The command %r failed with exit code %d.\nError:\n%s\nOutput:\n%s\n" % (
            cmd, returncode, indent_block(stderr), indent_block(stdout))
        raise RunCommandException(msg, returncode, stdout, stderr)
    if print_output:
        print(stdout)
    return stdout


def parse_version(x: str):
    return tuple(int(_) for _ in x.split("."))


def render_version(t: Tuple[int, int, int]):
    return ".".join(str(_) for _ in t)


def load_class(name: str, module: str, package: Optional[str] = None) -> Any:
    logger.debug(f"Loading class {package or ''}.{module}.{name}")
    mod = importlib.import_module(name=module, package=package)
    return getattr(mod, name)


def provider_username_project_from_git_url(url: str) -> Tuple[str, str, str]:
    ssh_pattern = "git@([^:]+):([^/]+)/(.+)"
    res = re.search(ssh_pattern, url, re.IGNORECASE)
    if res:
        return res.group(1), res.group(2), res.group(3)
    https_pattern = "https://([^/]+)/([^/]+)/(.+)"
    res = re.search(https_pattern, url, re.IGNORECASE)
    if res:
        return res.group(1), res.group(2), res.group(3)


def ln(s: str) -> int:
    """
    Computes the length of a string while taking into account the coloring characters.
    """
    return len(re.sub(rb"(\x1b|\[\d{1,2}m)*", b"", s.encode("utf-8")))


def text_distribute(chunks: List[str], width: int) -> str:
    remaining: int = width - sum(map(ln, chunks))
    spaces: List[int] = [0] + [int(floor(remaining / (len(chunks) - 1)))] * (len(chunks) - 1)
    spaces[-1] += remaining - sum(spaces)

    txt: str = ""
    for s, t in zip(spaces, chunks):
        txt += f"{' ' * s}{t}"
    return txt


def text_justify(txt: str, width: int) -> str:
    return "\n".join(
        [" " * int(floor((width - ln(line)) / 2)) + line for line in txt.splitlines(keepends=False)]
    )


def validator_token(token: str) -> bool:
    try:
        DuckietownToken.from_string(token, allow_expired=False)
    except dt_authentication.exceptions.InvalidToken as e:
        # logger.error(f"The given token is not valid: {str(e)}")
        return False
    except dt_authentication.exceptions.ExpiredToken:
        # logger.warning(f"The given token is expired, make sure you get a fresh one before continuing.")
        return False
    # ---
    return True


def yellow_bold(x: Any) -> str:
    return colored(str(x), color="yellow", attrs=["bold"])


class DebugInfo:
    name2versions: Dict[str, Union[str, Dict[str, str]]] = {}


def pip_install(interpreter: str, requirements: str):
    try:
        subprocess.check_output(
            [interpreter, "-m", "pip", "install", "-r", requirements],
            stderr=subprocess.STDOUT,
        )
    except subprocess.CalledProcessError as e:
        msg: str = "An error occurred while installing python dependencies"
        raise ShellInitException(msg, stdout=e.stdout, stderr=e.stderr)


def indent_block(s: str, indent_len: int = 4) -> str:
    space: str = " " * indent_len
    return space + f"\n{space}".join(s.splitlines() if s is not None else ["None"])


def pretty_json(data: Any, indent_len: int = 0) -> str:
    return indent_block(json.dumps(data, sort_keys=True, indent=4), indent_len=indent_len)


def pretty_exc(exc: Exception, indent_len: int = 0) -> str:
    return indent_block(
        ''.join(traceback.TracebackException.from_exception(exc).format()), indent_len=indent_len)


def ensure_bash_completion_installed():
    import dt_shell
    from dt_shell.database import DTShellDatabase
    if platform.system() in ["Linux", "Darwin"]:
        db: DTShellDatabase = DTShellDatabase.open("bash-completion-install")
        key: str = f"dts-comletion-{dt_shell.__version__}"
        if not db.contains(key):
            logger.info("Installing bash-completion script...")
            src: str = os.path.join(SHELL_LIB_DIR, "assets", "dts-completion.bash")
            dst: str = os.path.join(BASH_COMPLETION_DIR, "dts")
            try:
                shutil.copyfile(src, dst)
            except Exception:
                logger.warning("An error occurred while attempting to install the bash-completion script")
                if DTShellConstants.VERBOSE:
                    traceback.print_last()
                    return
            # mark it as installed in the database
            db.set(key, True)
            logger.info("Bash-completion script successfully installed")
        else:
            logger.debug("Bash-completion script marked as already installed")


def print_debug_info() -> None:
    from .logging import dts_print
    from .checks.packages import _get_installed_distributions

    v = DebugInfo.name2versions
    v["python"] = sys.version
    v["duckietown-shell"] = __version__

    v["encodings"] = {
        "stdout": sys.stdout.encoding,
        "stderr": sys.stderr.encoding,
        "locale": locale.getpreferredencoding(),
    }

    try:
        installed = _get_installed_distributions()
        pkgs = {_.project_name: _.version for _ in installed}
        for pkg_name, pkg_version in pkgs.items():
            include = (
                    ("duckietown" in pkg_name)
                    or ("dt-" in pkg_name)
                    or ("-z" in pkg_name)
                    or ("aido" in pkg_name)
            )
            if include:
                v[pkg_name] = pkg_version
    except (ImportError, AttributeError):
        logger.warning('Please update "pip" to have better debug info.')

    versions = yaml.dump(v, default_flow_style=False)
    fn = "~/shell-debug-info.txt"
    fn = os.path.expanduser(fn)
    with open(fn, "w") as f:
        f.write(versions)
    msg = f"To report a bug, please also include the contents of {fn}"
    dts_print(msg, "yellow")


def env_option(key: str, default: Any = NOTSET, true_choices: List[str] = None) -> Optional[Any]:
    if default in [NOTSET, True, False]:
        # boolean options
        true_choices = true_choices or ["1", "true", "yes"]
        default = "1" if default is True else "0"
        return os.environ.get(key, default).lower().strip() in true_choices
    else:
        # non-boolean options
        if true_choices is not None:
            raise ValueError("You cannot use 'default' and 'true_choices' at the same time.")
        return os.environ.get(key, default)



import builtins
import json
import sys
from typing import List

from dt_authentication import DuckietownToken, InvalidToken
from dt_shell import DTCommandAbs, DTShell, dtslogger


class DTCommand(DTCommandAbs):
    @staticmethod
    def command(shell: DTShell, args: List[str]):
        try:
            if args:
                token_s = args[0]
            else:
                msg = "Please enter token:\n> "
                token_s = builtins.input(msg)

            dtslogger.debug("Verifying token %r\n" % token_s)

            try:
                token = DuckietownToken.from_string(token_s)
            except InvalidToken:
                msg = "Invalid token format."
                dtslogger.error(msg + "\n")
                sys.exit(3)

            dtslogger.info("Token parsed correctly!")
            payload = json.dumps(token.payload)
            print(f"\nToken content:\n{payload}")
            sys.exit(0)

        except Exception as e:
            dtslogger.error(str(e) + "\n")
            sys.exit(3)



import os
import sys

from setuptools import setup, find_packages


def get_version(filename):
    import ast
    version = None
    with open(filename) as f:
        for line in f:
            if line.startswith('__version__'):
                version = ast.parse(line).body[0].value.s
                break
        else:
            raise ValueError('No version found in %r.' % filename)
    if version is None:
        raise ValueError(filename)
    return version


if sys.version_info < (3, 6):
    msg = 'duckietown-shell works with Python 3.6 and later.\nDetected %s.' % str(sys.version)
    sys.exit(msg)

distro = 'daffy'

shell_version = get_version(filename='lib/dt_shell/__init__.py')

shell_requires = [
    "termcolor>=2.3.0,<3",
    "requests>=2.31.0,<3",
    # force pyyaml away from specific versions: https://github.com/yaml/pyyaml/issues/724
    "pyyaml!=6.0.0,!=5.4.0,!=5.4.1,<7",

    # CLI utils
    "pyfiglet>=0.8.0,<1",
    "questionary>=2.0.1,<3",
    "argcomplete>=3.1.4,<4",

    # database concurrent access control
    "filelock>=3,<4",

    # duckietown deps
    "dockertown>=0.2.2,<1",
    "dtproject>=1.0.6,<2",
    "dt-authentication>=2.1.4,<3",
    # TODO: this cannot be released until DCSS is released on the HUB
    'dt-data-api>=2.0.0,<3',
]

compatibility_requires = [
    # NOTE: this is used by checks/environment.py and cannot be removed for compatibility
    "docker>=6.1.3,<7",
]

install_requires = shell_requires + compatibility_requires

system_version = tuple(sys.version_info)[:3]
if system_version < (3, 7):
    install_requires.append('dataclasses')


# embed requirements as asset into the release
requirements_fpath: str = os.path.join(
    os.path.dirname(__file__), "lib", "dt_shell", "assets", "requirements.txt"
)
with open(requirements_fpath, "wt") as fout:
    fout.write("\n".join(shell_requires))


setup(
    name='duckietown-shell',
    version=shell_version,
    download_url='http://github.com/duckietown/duckietown-shell/tarball/%s' % shell_version,
    package_dir={
        'dt_shell': 'lib/dt_shell',
        'dt_shell_cli': 'lib/dt_shell_cli',
    },
    packages=find_packages("lib", exclude=["dt_shell_tests"]),
    # we want the python 2 version to download it, and then exit with an error
    # python_requires='>=3.6',

    tests_require=[],
    install_requires=install_requires,
    # This avoids creating the egg file, which is a zip file, which makes our data
    # inaccessible by dir_from_package_name()
    zip_safe=False,

    # without this, the stuff is included but not installed
    include_package_data=True,
    package_data={
        'dt_shell': [
            'embedded/*',
            'embedded/*/*',
            'embedded/*/*/*',
            'embedded/*/*/*/*',
            'assets/*',
            'assets/*/*',
            'assets/*/*/*',
            'assets/*/*/*/*',
        ],
        'dt_shell_cli': [],
    },

    entry_points={
        'console_scripts': [
            'dts = dt_shell_cli.dts:dts',
        ]
    }
)



import copy
from dataclasses import dataclass
from typing import Dict, Optional


@dataclass
class ShellConfig:
    token_dt1: Optional[str]  # key
    docker_username: Optional[str]
    docker_password: Optional[str]
    duckietown_version: Optional[str]  # ente, daffy, master19, ...
    docker_credentials: Dict[str, Dict[str, str]]  # username, secret


_instance: Optional[ShellConfig] = None


def get_shell_config_default() -> ShellConfig:
    return ShellConfig(
        token_dt1=None,
        docker_username=None,
        docker_password=None,
        duckietown_version=None,
        docker_credentials={},
    )


def read_shell_config(shell) -> ShellConfig:
    global _instance
    if _instance is None:
        docker_credentials: Dict[str, dict] = {
            registry: copy.deepcopy(credentials)
            for registry, credentials in shell.profile.secrets.docker_credentials.items()
        }
        # rename password -> secret
        for credentials in docker_credentials.values():
            credentials["secret"] = credentials.pop("password")
        # package as ShellConfig
        _instance = ShellConfig(
            token_dt1=shell.profile.secrets.dt1_token,
            docker_username=None,
            docker_password=None,
            duckietown_version=shell.profile.name,
            docker_credentials=docker_credentials,
        )
    return _instance



# used by the following commands files:
#
#   - start_gui_tools/command.py
#   - build_utils/command.py
#   - chalenges/command.py
#   - code/workbench/command.py
#   - duckiebot/calibrate_intrinsics/command.py
#   - duckiebot/demo/command.py
#   - duckiebot/evaluate/command.py
#   - duckiebot/keyboard_control/command.py
#   - duckiebot/led_control/command.py
#   - excercises/build/command.py
#   - exerrcises/test/command.py
#   - utils/docker_utils.py
#
def check_docker_environment():
    from dt_shell.checks.environment import check_docker_environment
    return check_docker_environment()


# used by the following commands files:
#
#   - build_utils/command.py
#   - chalenges/command.py
#
def check_package_version(*args, **kwargs):
    pass



from typing import List

import yaml

from dt_shell import __version__, DTCommandAbs, DTShell


class DTCommand(DTCommandAbs):
    help = "Prints out the version of the shell and returns."

    @staticmethod
    def command(shell: DTShell, args: List[str]):
        # noinspection PyDictCreation
        versions: dict = {}
        # shell version
        versions["shell"] = f"v{__version__}"
        # command sets
        for cs in shell.command_sets:
            command_set_versions: dict = versions.get("command sets", {})
            head_version, closest_version = cs.version, cs.closest_version
            # if we have a HEAD version, we use that one
            if head_version:
                command_set_versions[cs.name] = f"v{head_version}"
            # revert to the closest tag otherwise
            elif closest_version:
                command_set_versions[cs.name] = f"devel (closest: {closest_version})"
            else:
                command_set_versions[cs.name] = "NA"

            versions["command sets"] = command_set_versions
        # print versions
        print()
        print(yaml.safe_dump(versions, sort_keys=False).replace("'", ""))



import webbrowser
from typing import List

import termcolor
from dt_authentication import InvalidToken, DuckietownToken

from dt_shell import DTCommandAbs, dtslogger

from dt_shell import DTShell


class DTCommand(DTCommandAbs):
    @staticmethod
    def command(shell: DTShell, args: List[str]):
        link = "https://hub.duckietown.com/token"
        example = "dt2-7vEuJsaxeXXXXX-43dzqWFnWd8KBa1yev1g3UKnzVxZkkTbfSJnxzuJjWaANeMf4y6XSXBWTpJ7vWXXXX"
        msg = """
Please enter your Duckietown token.

It looks something like this:

    {example}

To find your token, first login to duckietown.org, and open the page:

    {link}


Enter token: """.format(
            link=href(link), example=dark(example)
        )

        shell.sprint("args: %s" % args.__repr__())

        if args:
            val_in = args[0]
        else:
            webbrowser.open(link, new=2)
            val_in = input(msg)

        s = val_in.strip()
        try:
            token = DuckietownToken.from_string(s)
            if token.uid == -1:
                msg = "This is the sample token. Please use your own token."
                raise ValueError(msg)
            shell.sprint("Correctly identified as uid = %s" % token.uid)
        except InvalidToken as e:
            msg = 'The string "%s" does not look like a valid token:\n%s' % (s, e)
            shell.sprint(msg)
            return False

        # make sure this token is of a version we support
        tokens_supported: List[str] = shell.profile.distro.tokens_supported
        if token.version not in tokens_supported:
            dtslogger.error(f"Token version '{token.version}' not supported by this profile's distro. "
                            f"Only versions supported are {tokens_supported}.")
            return False

        # update the profile token
        shell.profile.secrets.dt_token = s


def dark(x):
    return termcolor.colored(x, attrs=["dark"])


def href(x):
    return termcolor.colored(x, "blue", attrs=["underline"])



import argparse
from typing import Optional, List

from dt_shell.commands import DTCommandConfigurationAbs


class DTCommandConfiguration(DTCommandConfigurationAbs):

    @classmethod
    def parser(cls, **kwargs) -> Optional[argparse.ArgumentParser]:
        parser: argparse.ArgumentParser = argparse.ArgumentParser()
        parser.add_argument("profile", nargs="?", default=None, help="Name of the profile to switch to")
        # ---
        return parser

    @classmethod
    def aliases(cls) -> List[str]:
        return []
